package org.netkernel.layer0.meta.impl;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.ParserConfigurationException;

import org.netkernel.container.ILogger;
import org.netkernel.container.IMessages;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.GrammarParseException;
import org.netkernel.layer0.bnf.IGrammarGroup;
import org.netkernel.layer0.meta.IEndpointArgumentMeta;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.representation.IRepAny;
import org.netkernel.layer0.urii.ActiveURIEncoding;
import org.netkernel.layer0.urii.ParsedIdentifierImpl;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.urii.INetKernelException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class MetadataUtils
{
	
	public static IEndpointArgumentMeta getEndpointIdentifierPart(IEndpointMeta aMeta, String aName)
	{	IEndpointArgumentMeta result=null;
		IEndpointArgumentMeta[] parts=aMeta.getArguments();
		for (int i=0; i<parts.length; i++)
		{	if (parts[i].getName().equals(aName))
			{	result=parts[i];
				break;
			}
		}
		return result;
	}
	
	public static boolean match(IEndpointMeta aEndpointMeta, INKFRequestReadOnly aRequest)
	{	boolean result=false;
		int verb=aRequest.getVerb();
		if ((verb&aEndpointMeta.getSupportedVerbs())!=0) // verb matches
		{	if (verb==INKFRequestReadOnly.VERB_TRANSREPRESENT)
			{	//inbound and outbound representations tested
			
				Class[] responseClasses=aEndpointMeta.getResponseClasses();
				Class requestedClass = aRequest.getRepresentationClass();
				
				int i;
				if (responseClasses.length>1 && responseClasses[0]==IRepAny.class)
				{	//special case weak match
					for (i=1; i<responseClasses.length; i++)
					{	if (responseClasses[i].isAssignableFrom(requestedClass))
						{	break;
						}
					}
				}
				else
				{	for (i=0; i<responseClasses.length; i++)
					{	if (requestedClass.isAssignableFrom(responseClasses[i]))
						{	break;
						}
					}
				}
				if (responseClasses.length==0 || i<responseClasses.length)
				{	Object primary=aRequest.getPrimary();
					if (primary!=null)
					{	Class from = primary.getClass();
						IEndpointArgumentMeta parts[] = aEndpointMeta.getArguments();
						for (int j=0; j<parts.length; j++)
						{	Class fromClasses[] = parts[j].getRequiredRepresentations();
							for (i=0; i<fromClasses.length; i++)
							{	if (fromClasses[i].isAssignableFrom(from))
								{	result = true;
								}
							}
						}
					}
				}
			}
			else
			{	// other requests just look at identifier
				result = aEndpointMeta.getIdentifierGrammar().matches(aRequest.getIdentifier());
			}
		}
	
		return result;
	}

	public static boolean match(IEndpointMeta aEndpointMeta, INKFRequestReadOnly aRequest, INKFResolutionContext aContext)
	{	return match(aEndpointMeta,aRequest, aContext, null);
	}

	public static boolean match(IEndpointMeta aEndpointMeta, INKFRequestReadOnly aRequest, INKFResolutionContext aContext, String aEndpointId)
	{	boolean result=false;
		int verb=aRequest.getVerb();
		boolean transreptResponseMatch=false;
		if ((verb&aEndpointMeta.getSupportedVerbs())!=0) // verb matches
		{	if (verb==INKFRequestReadOnly.VERB_TRANSREPRESENT)
			{	//inbound and outbound representations tested
			
				Class[] responseClasses=aEndpointMeta.getResponseClasses();
				Class requestedClass = aRequest.getRepresentationClass();
				
				int i;
				if (responseClasses.length>1 && responseClasses[0]==IRepAny.class)
				{	//special case weak match
					for (i=1; i<responseClasses.length; i++)
					{	if (responseClasses[i].isAssignableFrom(requestedClass))
						{	break;
						}
					}
				}
				else
				{	for (i=0; i<responseClasses.length; i++)
					{	if (requestedClass.isAssignableFrom(responseClasses[i]))
						{	break;
						}
					}
				}
				if (responseClasses.length==0 || i<responseClasses.length)
				{	transreptResponseMatch=true;
					Object primary=aRequest.getPrimary();
					if (primary!=null)
					{	Class from = primary.getClass();
						IEndpointArgumentMeta parts[] = aEndpointMeta.getArguments();
						for (int j=0; j<parts.length; j++)
						{	Class fromClasses[] = parts[j].getRequiredRepresentations();
							for (i=0; i<fromClasses.length; i++)
							{	if (fromClasses[i].isAssignableFrom(from))
								{	result = true;
								}
							}
						}
					}
				}
			}
			else
			{	// other requests just look at identifier
				result = aEndpointMeta.getIdentifierGrammar().matches(aRequest.getIdentifier());
			}
		}
		
		if (aContext.isDebugResolve())
		{	
			ILogger l=aContext.getKernelContext().getKernel().getLogger();
			if (result)
			{	aContext.pushResolutionDebug();
				//NKFResolutionDebugEntity rde=new NKFResolutionDebugEntity(aContext.getResolution().getEndpoint(), aContext.getResolution().getScope());
				aContext.addResolutionDebug(l.format("MSG_LR0_RES_LOGICAL_MATCH",aEndpointMeta.getName()),0,aEndpointId);
				aContext.popResolutionDebug();
			}
			else
			{	String msg;
				if ((verb&aEndpointMeta.getSupportedVerbs())==0)
				{	if (verb!=INKFRequestReadOnly.VERB_TRANSREPRESENT && aEndpointMeta.getIdentifierGrammar().matches(aRequest.getIdentifier()))
					{	msg="verb (grammar matches)";
					}
					else
					{	msg="verb";
					}
				}
				else if (verb==INKFRequestReadOnly.VERB_TRANSREPRESENT)
				{	if (transreptResponseMatch)
					{	msg="primary";
					}
					else
					{	msg="response representation";
					}
				}
				else
				{	msg="grammar";
				}
				aContext.pushResolutionDebug();
				//NKFResolutionDebugEntity rde=new NKFResolutionDebugEntity(aContext.getResolution().getEndpoint(), aContext.getResolution().getScope());
				aContext.addResolutionDebug(l.format("MSG_LR0_RES_LOGICAL_NOMATCH",aEndpointMeta.getName(),msg),0,aEndpointId);
				aContext.popResolutionDebug();
			}
		}
		return result;
	}
	
	private static Map<String,IGrammarParser> sCustomGrammarParsersByTag=new HashMap<String,IGrammarParser>();
	static
	{	sCustomGrammarParsersByTag.put("active", new ActiveGrammarParser());
		sCustomGrammarParsersByTag.put("simple", new SimpleGrammarParser());
	}
	
	public static void parseGrammar(Element aGrammarDefn, EndpointMetaBuilder aBuilder, IMessages aMessages) throws ParserConfigurationException, GrammarParseException, INetKernelException
	{
		IGrammarParser customGrammar=null;
		Element ce = XMLUtils.getFirstChildElement(aGrammarDefn);
		if (ce!=null && XMLUtils.getNextSiblingElement(ce)==null)
		{	String tagName=XMLUtils.pleaseDOMGiveMeTheNameOf(ce);
			customGrammar=sCustomGrammarParsersByTag.get(tagName);
		}
		
		if (customGrammar!=null)
		{	customGrammar.parseGrammar(ce,aBuilder,aMessages);
		}
		else
		{	// low level grammar definition
			Document d=XMLUtils.newDocument();
			Element de = d.createElement("group");
			d.appendChild(de);
			Element e2=(Element)d.importNode(aGrammarDefn, true);
			Node n=e2.getFirstChild();
			while ( n!=null)
			{	de.appendChild(n);
				n=e2.getFirstChild();
			}
			IIdentifierGrammar grammar= BNFGrammarFactory.parse(de,aMessages);
			for (Element e=de; e!=null; e=XMLUtils.inOrderTraversalNext(e,de))
			{	String name=e.getAttribute("name");
				if (name.length()>0)
				{	String elementName=XMLUtils.pleaseDOMGiveMeTheNameOf(e);
					if (!elementName.equals("regex"))
					{
						String desc=e.getAttribute("desc");
						if (desc.length()==0) desc=null;
						Element firstChild=null;
						if (XMLUtils.getText(e, true, true).length()==0)
						{	for (Element child=XMLUtils.getFirstChildElement(e); child!=null; child=XMLUtils.getNextSiblingElement(child))
							{	if (firstChild!=null)
								{	firstChild=null;
									break;
								}
								else
								{	if (XMLUtils.pleaseDOMGiveMeTheNameOf(child).equals("regex"))
									{	firstChild=child;
									}
								}
							}
						}
						String type="unknown";
						if (firstChild!=null)
						{	String ta=firstChild.getAttribute("type");
							if (ta.length()>0)
							{	type=ta;
							}
						}
						IEndpointArgumentMeta argMeta;
						if (type.equals(BNFGrammarFactory.BUILTIN_ACTIVE_ESCAPED_URI) || type.equals(BNFGrammarFactory.BUILTIN_ACTIVE_ESCAPED_URI_SIMPLE))
						{	argMeta = new SourcedArgumentMetaImpl(name, desc, null, null);
						}
						else
						{	argMeta = new SubstringArgumentMetaImpl(name, desc, type);
						}
						
						aBuilder.addArgument(argMeta);			
					}
				}
				
			}
			aBuilder.setGrammar(grammar);
		}		
	}
	
	static IGrammarGroup createVarargs(IGrammarGroup aParent)
	{	IGrammarGroup g1 = BNFGrammarFactory.createGroup(aParent,0,Integer.MAX_VALUE);
		BNFGrammarFactory.createText(g1, "+");
		IGrammarGroup g2 = BNFGrammarFactory.createGroup(g1,BNFGrammarFactory.ARG_UNDEFINED_NAME,null);
		BNFGrammarFactory.createRegex(g2, BNFGrammarFactory.BUILTIN_NMTOKEN);
		BNFGrammarFactory.createText(g1, "@");
		g2 = BNFGrammarFactory.createGroup(g1,BNFGrammarFactory.ARG_UNDEFINED_VALUE, ActiveURIEncoding.INSTANCE);
		BNFGrammarFactory.createRegex(g2, BNFGrammarFactory.BUILTIN_ACTIVE_ESCAPED_URI_SIMPLE);
		return g1;
	}
	
	static IGrammarGroup createArgument(IGrammarGroup aParent, String aName, int aMin, int aMax)
	{	IGrammarGroup g0 = BNFGrammarFactory.createGroup(aParent, aMin, aMax);
		BNFGrammarFactory.createText(g0, "+"+aName+"@");
		IGrammarGroup g1 = BNFGrammarFactory.createGroup(g0, aName, ActiveURIEncoding.INSTANCE);
		BNFGrammarFactory.createRegex(g1, BNFGrammarFactory.BUILTIN_ACTIVE_ESCAPED_URI_SIMPLE);
		return g0;
	}
	
	/*
	Below is an unused implementation that supports regex matching on the active URI argument
	Not used to avoid accidental breaks of pbv: etc
	
	private static final String ESCAPED_PLUS_ENCODING="EsCpLuS";

	static IGrammarGroup createArgument(IGrammarGroup aParent, String aName, int aMin, int aMax, String regex, boolean regexraw)
	{	IGrammarGroup g0 = BNFGrammarFactory.createGroup(aParent, aMin, aMax);
		BNFGrammarFactory.createText(g0, "+"+aName+"@");
		//Original code
		IGrammarGroup g1 = BNFGrammarFactory.createGroup(g0, aName, ActiveURIEncoding.INSTANCE);
		BNFGrammarFactory.createRegex(g1, BNFGrammarFactory.BUILTIN_ACTIVE_ESCAPED_URI_SIMPLE);
		if(regex==null)
		{
			IGrammarGroup g1 = BNFGrammarFactory.createGroup(g0, aName, ActiveURIEncoding.INSTANCE);
			BNFGrammarFactory.createRegex(g1, BNFGrammarFactory.BUILTIN_ACTIVE_ESCAPED_URI_SIMPLE);
		}
		else
		{	//Optional unused support for regex patterns on the argument.  !regexraw does an active URI encoding of the regex itself.
			IGrammarGroup g1 = BNFGrammarFactory.createGroup(g0, aName, null);
			if(!regexraw)
			{
				regex=regex.replaceAll("\\x2B", ESCAPED_PLUS_ENCODING);
				regex=ActiveURIEncoding.INSTANCE.encode(regex);
				regex=regex.replaceAll(ESCAPED_PLUS_ENCODING, "+");
			}
			BNFGrammarFactory.createRegex(g1, regex);
		}
		
		return g0;
	}
	*/
}
