package org.netkernel.layer0.nkf.impl;

import org.netkernel.layer0.meta.*;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.urii.ParsedIdentifierImpl;
import org.netkernel.layer0.util.OrderedCheapMultiStringMap;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.IResponseMeta;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.request.impl.RequestImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IIdentifier;


public class NKFEndpointRequestImpl extends NKFRequestImpl
{
	private IEndpointMeta mEndpointMeta;
	
	public NKFEndpointRequestImpl(NKFContextImpl aContext, IEndpointMeta aEndpointMeta) 
	{	super(aContext,null);
		mEndpointMeta=aEndpointMeta;
	}
	
	protected void addArgument(String aName, String aURI, Object aRepresentation, IResponseMeta aMeta)
	{	
		IEndpointArgumentMeta argMeta=MetadataUtils.getEndpointIdentifierPart(mEndpointMeta, aName);
		boolean notIdentifier=argMeta!=null && argMeta.getArgumentType()==IEndpointArgumentMeta.ArgumentType.STRING;
		if (aRepresentation!=NO_REP && notIdentifier)
		{	throw createIAE("EX_NKF_REQ_ARG_PBV_NOT_URI",aName);
			//System.out.println("addArgument "+aName+" "+aURI);
		}
		super.addArgument(aName, aURI, aRepresentation, aMeta);
		
	}
	public void addArgument(String aName, String aURI) throws NKFException
	{	if (aName==null) throw createIAE("EX_NKF_REQ_ARG_NAME");
		if (aURI==null) throw createIAE("EX_NKF_REQ_ARG_URI");
		try
		{	
			IEndpointArgumentMeta argMeta=MetadataUtils.getEndpointIdentifierPart(mEndpointMeta, aName);
			boolean isIdentifier=argMeta!=null && argMeta.getArgumentType()!=IEndpointArgumentMeta.ArgumentType.STRING;
			String uri = mContext.resolveIdentifier(aURI,isIdentifier);
			
			if (!isIdentifier)
			{	if (!mEndpointMeta.getIdentifierGrammar().partMatches(aName, uri))
				{	NKFException e2=mContext.createNKFE("EX_NKF_EPREQ_BAD_ARG", "MSG_NKF_EPREQ_BAD_ARG", aName,uri,mEndpointMeta.getName());
					throw e2;
				}
			}
			
			addArgument(aName, uri, NO_REP, null);
		}
		catch (LiteralParameterException e)
		{	Object parameterRepresentation = e.getRepresentation();
			this.addArgumentByValue(aName, parameterRepresentation);
		}
	}	
	
	public IRequest getKernelRequest() throws NKFException
	{	
		IRequestScopeLevel scope=null;
		if (mScope!=null)
		{	scope=mScope;
		}
		else
		{	scope=mContext.getRequestScope();
		}
		
		IRequest parentRequest = mContext.getThisKernelRequestOrNull();
		
		// construct request
		IIdentifier id=innerGetIdentifier();
		if (mValueSpace!=null)
		{	scope = RequestScopeLevelImpl.createDurableScopeLevel(mValueSpace,scope);
			try
			{	mValueSpace.onCommissionSpace(mContext.getKernel());
			} catch (Exception e)
			{	// won't happen
			}
		}
		
		Object[] values=null;
		if (mPrimary!=null)
		{	values = new Object[] { mPrimary };
		}
		
		IRequestResponseFields headers=determineHeaders();
		
		return new RequestImpl(mType,id,scope,mRepresentationClass,parentRequest,null,headers,values);
	}
	
	public String getIdentifier() throws NKFException
	{	return innerGetIdentifier().toString();
	}
	
	private IIdentifier innerGetIdentifier() throws NKFException
	{	try
		{	if (mArguments==null)
			{	mArguments = new OrderedCheapMultiStringMap(2);
			}
			IIdentifierGrammar grammar = mEndpointMeta.getIdentifierGrammar();
			return ParsedIdentifierImpl.parseGrammar(mArguments,grammar,mContext.getKernel().getLogger());
		}
		catch (GrammarException e)
		{	NKFException e2=mContext.createNKFE("EX_NKF_EPREQ_MISSING_ARGS", "MSG_NKF_EPREQ_MISSING_ARGS", mEndpointMeta.getName(),e.getMessage());
			throw e2;
		}		
	}

}
