/******************************************************************************
 * (c) Copyright 2002,2003, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.urii;


import org.netkernel.container.ILogger;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.util.Utils;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;

import java.io.*;
import java.net.*;
import java.util.concurrent.atomic.AtomicReference;

/**
 * The representation returned by the classloader resource endpoint. It expires if the resource is a file and it changes
 * @author  tab
 */
public final class ClassLoaderResourceRepresentation extends ClassLoaderResourceExistence 
	implements IReadableBinaryStreamRepresentation, IBinaryStreamRepresentation
{
	private AtomicReference<URLConnection> mConnection;
	private int mLength=-1;
	private String mEncoding;
	private String mMime;
	
	private static boolean sOptimizeFileHandles=false;
	
	public static void setOptimizeFileHandles(boolean aOptimize)
	{	sOptimizeFileHandles=aOptimize;
	}
	
	/** Creates a new instance of ClassLoaderResourceRepresentation
	 * @param aPath path of the resource
	 * @param aClassLoader classloader to find the resource
	 * @param aLogger logger to format any exceptions
	 * @param aContext context (used for debug)
	 * @param aExpiryOffset the period between checks for expiry (so as not to overload IO)
	 * @throws IOException thrown if there is a problem accessing resource
	 * @throws NetKernelException thrown for higher level problems
	 */
	public ClassLoaderResourceRepresentation(String aPath, ClassLoader aClassLoader) throws IOException
	{	
		super(aPath,aClassLoader);
		
		URL url=getURL();
		if (url!=null)
		{	URLConnection c=mURL.openConnection();
			mMime = c.getContentType();
			mLength = c.getContentLength();
			if(!mURL.getProtocol().equals("jar"))
			{	mEncoding = c.getContentEncoding();
			}
			if (sOptimizeFileHandles)
			{	mConnection=new AtomicReference<URLConnection>(c);
			}
			else
			{	c.getInputStream().close();
			}
		}
	}
	
	public static INetKernelException generateNotFoundException(ISpace aSpace, String aPath, ILogger aLogger)
	{	String spaceId;
		if (aSpace instanceof ISpaceWithIdentity && ((ISpaceWithIdentity)aSpace).getIdentifier()!=null)
		{	spaceId=((ISpaceWithIdentity)aSpace).getIdentifier().toString();
		}
		else
		{	spaceId = aSpace.toString();
		}
		return Utils.createFormattedException("EX_LR0_RES_NOTFOUND","MSG_LR0_RES",aLogger,"/"+aPath,spaceId);
	}
	
	/** return the length of the resource */
	public int getContentLength()
	{	return mLength;
	}
	/** write the resource to the given output stream */
	public void write(OutputStream aStream) throws IOException
	{	InputStream is = ensureOpenConnection().getInputStream();
		Utils.pipeNoClose(is,aStream);
	}
	/** get an input stream for the  resource */
	public InputStream getInputStream() throws IOException
	{	return ensureOpenConnection().getInputStream();
	}
	private URLConnection ensureOpenConnection() throws IOException
	{	URLConnection result=mConnection!=null ? mConnection.getAndSet(null) : null;
		if (result==null)
		{	result = mURL.openConnection();
		}
		return result;
	}
	
	/** returns a guess at the encoding if the resource is text
	 */
	public String getEncoding()
	{	return mEncoding;
	}
	
	public String getMime()
	{	return mMime;
	}	

}